@@ -353,3 +353,28 @@ mutation UpdateLock($id: ID!, $active: Boolean!, $endAt: DateTime) {
   }
 }
 '''
+
+update_context = '''
+mutation UpdateContext(
+  $queue: String!
+  $context: Json!
+){
+  updateQueue(
+    findBy:{ name: $queue }
+    queue: {
+      context: $context
+    }
+  ){
+    successful
+    messages{
+      field
+      message
+    }
+    result{
+      id
+      name
+      context
+    }
+  }
+}
+'''
