@@ -32,6 +32,7 @@ query WorkerJobs($worker: String){
           name
           alias
           type
+          context # for rehydration
           enabled
           schedule
           query
@@ -54,18 +55,6 @@ query WorkerWaitingJobs($worker: String){
     jobStatus: WAITING
   }) {
     id
-    input
-    output
-    progress
-    queue {
-      id
-      name
-      alias
-      type
-      enabled
-      schedule
-      query
-    }
   }
 }
 '''
@@ -121,6 +110,7 @@ all_queues = '''
     name
     alias
     description
+    context
     enabled
     type
     schedule
