@@ -3,7 +3,7 @@ import time
 import json
 from json.decoder import JSONDecodeError
 import multiprocessing as mp
-from multiprocessing import SimpleQueue
+from multiprocessing import SimpleQueue, Pipe
 from datetime import datetime, timedelta
 from enum import Enum
 from pygqlc import GraphQLClient
@@ -53,6 +53,7 @@ class ValiotWorker():
     self.queues = {}
     self.eventQueues = {}
     self.currentQueues = []
+    self.runningJobs = {}
     self.lockedJobs = []
     self.gql = GraphQLClient()
     self.worker = None
@@ -238,14 +239,29 @@ class ValiotWorker():
       safe_vars = {**variables, 'status': status.value}
       # this allows to specify ID inside job for backwards compatibility
       return uploaders.update_job({'id': job['id'], **safe_vars})
+    def update_this_context(context):
+      return self.gql.mutate(mutations.update_context, {
+        'queue': __.get(job, 'queue.name'),
+        'context': json.dumps(context)
+      })
+    src_status_conn, dst_status_conn = Pipe()
+    def get_job_status():
+      if dst_status_conn.poll():
+        job = dst_status_conn.recv()
+        return JobStatus(job['jobStatus'])
+      return None
     # ! Inyecta a los kwargs el job que se está ejecutando:
     kwargs['job'] = job
     kwargs['job_id'] = job['id']
     kwargs['job']['update'] = update_this_job # couple of aliases for the same fn
     kwargs['update_job'] = update_this_job  # couple of aliases for the same fn
+    kwargs['job']['get_status'] = get_job_status
+    kwargs['get_job_status'] = get_job_status
     # acceso directo a atributos del queue al que pertenece el job
     kwargs['queue'] = job['queue']
     kwargs['log'] = self.getProcessLoggerFunction(queue=job['queue'], job=job)
+    kwargs['context'] = __.get(job, 'queue.context', dict())
+    kwargs['update_context'] = update_this_context
     kwargs['kwargs'] = {**kwargs}
     variables = {
       'id': int(job['id']),
@@ -271,6 +287,12 @@ class ValiotWorker():
       p = self.context.Process(target=safe_fn, kwargs=(kwargs))
       self.log(LogLevel.INFO, f'started job with ID {job["id"]}')
       p.start()
+      self.runningJobs[job['id']] = {
+        'process': p,
+        'job': {**job},
+        'status_src': src_status_conn,
+        'status_dst': dst_status_conn
+      }
       if lock:
         self.lockedJobs.append({
           'process': p,
@@ -326,11 +348,12 @@ class ValiotWorker():
       description = __.get(queue, 'description', None)
       schedule = __.get(queue, 'schedule', None)
       query = __.get(queue, 'query', None)
-      lockRequired = __.get(queue, 'lockRequired', None)
+      lockRequired = __.get(queue, 'lockRequired', False)
       variables = {
         'name':queue['name'],
         'alias':queue['alias'],
         'type': queue['type'],
+        'lockRequired': lockRequired
       }
       # ! Add optional variables
       if enabled is not None:
@@ -341,8 +364,6 @@ class ValiotWorker():
         variables['schedule'] = schedule
       if query:
         variables['query'] = query
-      if lockRequired:
-        variables['lockRequired'] = lockRequired
       data, errors = self.gql.mutate(
         mutation=mutations.upsert_queue,
         variables=variables
@@ -353,15 +374,24 @@ class ValiotWorker():
       else:
         self.log(LogLevel.SUCCESS, f'Queue {name} actualizado correctamente')
 
+  def clearFinishedJobs(self):
+    finished = []
+    for job_id, running_job in self.runningJobs.items():
+      if not running_job['process'].is_alive():
+        finished.append(job_id)
+    for f in finished:
+      self.runningJobs.pop(f, None) # remove any finished jobs
+
   def unlockFinishedJobs(self, force=False):
     unlocked = []
+    _log = log if force else self.log
     for i, job_lock in enumerate(self.lockedJobs):
       if not job_lock['process'].is_alive() or force:
         if self.unlockQueue(job_lock['job']['queue']['name'], job_lock['lock']):
-          self.log(LogLevel.SUCCESS, f'{job_lock["job"]["queue"]["name"]} UNLOCKED')
+          _log(LogLevel.SUCCESS, f'{job_lock["job"]["queue"]["name"]} UNLOCKED')
           unlocked.append(i)
         else:
-          self.log(LogLevel.ERROR, f'error unlocking {job_lock["job"]["queue"]["name"]}')
+          _log(LogLevel.ERROR, f'error unlocking {job_lock["job"]["queue"]["name"]}')
     for u in unlocked:
       self.lockedJobs.pop(u)
 
@@ -471,12 +501,26 @@ class ValiotWorker():
       safe_vars = {**variables, 'status': status.value}
       # this allows to specify ID inside job for backwards compatibility
       return uploaders.update_job({'id': job['id'], **safe_vars})
+    def update_this_context(context):
+      return self.gql.mutate(mutations.update_context, {
+          'queue': __.get(currentQueue, 'name'),
+          'context': json.dumps(context)
+      })
+    src_status_conn, dst_status_conn = Pipe()
+    def get_job_status():
+      if dst_status_conn.poll():
+        job = dst_status_conn.recv()
+        return JobStatus(job['jobStatus'])
+      return None
     # ! Inyecta a los kwargs el job que se está ejecutando:
     kwargs = {
       'job_id': job['id'],
-      'job': {**job, 'update': update_this_job},
+      'job': {**job, 'update': update_this_job, 'get_status': get_job_status},
       'update_job': update_this_job,
+      'get_job_status': get_job_status,
       'queue': currentQueue,
+      'context': __.get(currentQueue, 'context', dict()),
+      'update_context': update_this_context,
       'data': data,
       'log': self.getProcessLoggerFunction(queue=currentQueue, job=job)
     }
@@ -492,6 +536,12 @@ class ValiotWorker():
         self.log(LogLevel.ERROR, f'\t...')
     p = self.context.Process( target=safe_fn, kwargs=(kwargs))
     p.start()
+    self.runningJobs[job['id']] = {
+        'process': p,
+        'job': {**job},
+        'status_src': src_status_conn,
+        'status_dst': dst_status_conn
+    }
     if lock:
       self.lockedJobs.append({
         'process': p,
@@ -529,6 +579,26 @@ class ValiotWorker():
       else:
         self.log(LogLevel.ERROR, f'ERROR REGISTERING LISTENER FOR EVENT QUEUE { queue["name"] }')
 
+  def registerJobUpdatesListener(self):
+    def onJobUpdated(data):
+      successful = __.get(data, 'successful')
+      job = __.get(data, 'result')
+      # check if update even succeeded
+      if not job or not successful:
+        return
+      # check if updated job matches any currently running
+      job_running = self.runningJobs.get(job['id'], None)
+      if not job_running:
+        return
+      # at this point we are certain the updated job is running, update structure to inform the job:
+      status_conn = self.runningJobs[job['id']]['status_src']
+      status_conn.send(job)
+    unsub = self.gql.subscribe(query=subscriptions.JOB_UPDATED, callback=onJobUpdated)
+    if (unsub):
+      self.log( LogLevel.SUCCESS, f'REGISTERED LISTENER FOR JOB UPDATES' )
+    else:
+      self.log( LogLevel.ERROR, f'ERROR REGISTERING LISTENER FOR JOB UPDATES' )
+
   def getProcessLoggerFunction(self, queue=None, job=None):
     """Logs to STDOUT formatted to help get helpful information
     
@@ -567,17 +637,20 @@ class ValiotWorker():
     self.log = self.getProcessLoggerFunction()
     # FN DEFINITION --------------------------
     def loggingLoop():
-      while not self.stopLogging:
-        # * check mailbox for every queue
-        # * log as many messages from same queue as possible
-        # TODO: May set a limit to this, in message count or timeout for other queues)
-        # ! Send messages to log-upload queue
-        # ! flag as CREATE or UPSERT for retention or real-time visualization only
-        for _, log_queue in self.logQueues.items():
-          while not log_queue.empty():
-            msg = log_queue.get()
-            log(msg['level'], msg['content'])
-      self.log(LogLevel.WARNING, 'Logging loop stopped')
+      try:
+        while not self.stopLogging:
+          # * check mailbox for every queue
+          # * log as many messages from same queue as possible
+          # TODO: May set a limit to this, in message count or timeout for other queues)
+          # ! Send messages to log-upload queue
+          # ! flag as CREATE or UPSERT for retention or real-time visualization only
+          for _, log_queue in self.logQueues.items():
+            while not log_queue.empty():
+              msg = log_queue.get()
+              log(msg['level'], msg['content'])
+      except KeyboardInterrupt:
+        log(LogLevel.WARNING, 'Logging stopped due to keyboard interrupt')
+      log(LogLevel.WARNING, 'Logging loop stopped')
     # END FN DEFINITION --------------------------
     # Create Queues for each queue:
     for queue_name, _ in self.queues.items():
@@ -597,6 +670,7 @@ class ValiotWorker():
         self.log(LogLevel.ERROR, f'errors: {errors}')
         return
       self.currentQueues = currentQueues
+      self.clearFinishedJobs()
       self.unlockFinishedJobs()
       self.runFrequencyQueues()
       self.runTimedOutEvents()
@@ -629,11 +703,12 @@ class ValiotWorker():
     if (self.pollingMode == PollingMode.SUBSCRIPTION):
       self.registerJobCreatedSubscription()
     self.registerEventJobs()
+    self.registerJobUpdatesListener()
     while 1:
       try:
         self.eventLoop(interval=interval)
       except KeyboardInterrupt:
-        self.log(LogLevel.WARNING, "\nuser's stop signal, exiting...")
+        log(LogLevel.WARNING, "\nuser's stop signal, exiting...")
         self.stopLoggingLoop()
         self.unlockFinishedJobs(force=True)
         self.gql.close()
